# Generated by Django 3.0.6 on 2020-12-19 22:45

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_scoped_permissions', '0002_scopedpermissiongroup'),
        ('tests', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='user',
            name='scoped_permission_groups',
            field=models.ManyToManyField(blank=True, to='django_scoped_permissions.ScopedPermissionGroup'),
        ),
        migrations.AddField(
            model_name='usertype',
            name='scoped_permission_groups',
            field=models.ManyToManyField(blank=True, to='django_scoped_permissions.ScopedPermissionGroup'),
        ),
    ]



# Generated by Django 3.0.6 on 2020-12-19 22:45

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_scoped_permissions', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='ScopedPermissionGroup',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.TextField()),
                ('scoped_permissions', models.ManyToManyField(blank=True, related_name='in_groups', to='django_scoped_permissions.ScopedPermission')),
            ],
        ),
    ]



from functools import wraps

from django.core.exceptions import PermissionDenied

from django_scoped_permissions.util import expand_scopes_from_context


def gql_has_scoped_permissions(
    *permissions, fail_message: str = "You are not permitted to view this",
):
    """
    gql_has_permissions is a function which wraps a `resolve_<x>` or
    `mutate` field for any GraphQL object.

    When called, it checks whether or not the calling user has permission to
    authorize the resource being requested, depending on the permissions given
    as necessary to access the resource.

    :param permissions: The permission required to access the wrapped resource.
    :param fail_to_none: If true, and the user is not authorized, the field will resolve to None.
                            If false, the entire query will fail dramatically in a 401.
    :param fail_message: If fail_to_none is false, and the permission fails, this variable determines
                         the string which is thrown in the exception.
    :param fail_to_lambda: If not none, and the permission fails, this variable (assumed to be a function)
                           will be called.
    :return:
    """

    def decorator(func):
        @wraps(func)
        def wrapper(cls, info, *args, **kwargs):
            if not hasattr(info, "context") or not hasattr(info.context, "user"):
                raise PermissionDenied(fail_message)

            context = {}
            context["context"] = info.context
            context["user"] = info.user
            expanded_permissions = expand_scopes_from_context(permissions, context)

            user = info.context.user
            if (
                (not user or user.is_anonymous)
                and len(expanded_permissions) > 0
                or not user.has_any_scoped_permissions(*expanded_permissions)
            ):
                raise PermissionDenied(fail_message)
            return func(cls, info, *args, **kwargs)

        return wrapper

    return decorator


gql_has_any_scoped_permissions = gql_has_scoped_permissions


def gql_has_all_scoped_permissions(
    *permissions, fail_message: str = "You are not permitted to view this",
):
    """
    gql_has_permissions is a function which wraps a `resolve_<x>` or
    `mutate` field for any GraphQL object.

    When called, it checks whether or not the calling user has permission to
    authorize the resource being requested, depending on the permissions given
    as necessary to access the resource.

    :param permissions: The permission required to access the wrapped resource.
    :param fail_to_none: If true, and the user is not authorized, the field will resolve to None.
                            If false, the entire query will fail dramatically in a 401.
    :param fail_message: If fail_to_none is false, and the permission fails, this variable determines
                         the string which is thrown in the exception.
    :param fail_to_lambda: If not none, and the permission fails, this variable (assumed to be a function)
                           will be called.
    :return:
    """

    def decorator(func):
        @wraps(func)
        def wrapper(cls, info, *args, **kwargs):
            if not hasattr(info, "context") or not hasattr(info.context, "user"):
                raise PermissionDenied(fail_message)

            context = {}
            context["context"] = info.context
            context["user"] = info.user
            expanded_permissions = expand_scopes_from_context(permissions, context)

            user = info.context.user
            if (
                (not user or user.is_anonymous)
                and len(permissions) > 0
                or not user.has_all_scoped_permissions(*expanded_permissions)
            ):
                raise PermissionDenied(fail_message)
            return func(cls, info, *args, **kwargs)

        return wrapper

    return decorator



from django.contrib.auth.models import AbstractUser
from django.db import models

from django_scoped_permissions.core import create_scope
from django_scoped_permissions.models import ScopedPermissionHolder, ScopedModelMixin


class User(ScopedPermissionHolder, AbstractUser):
    class Meta:
        indexes = (models.Index(fields=("email",)),)

    email = models.EmailField(unique=True)

    has_registered = models.BooleanField(default=False)

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    def get_granting_scopes(self):
        scopes = self.resolved_scopes

        for user_type in self.user_types.all():
            scopes.extend(user_type.get_granting_scopes())

        scopes.append(f"user:{self.id}")
        return scopes

    def get_name_or_username(self):
        return self.get_full_name() or self.username

    def clean(self):
        self.username = self.username.lower()
        self.email = self.email.lower()

    def __str__(self):
        return "User %s" % (self.username,)


class Company(models.Model):
    class Meta:
        pass

    name = models.CharField(max_length=128)
    short_name = models.CharField(max_length=32)

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    users = models.ManyToManyField(User, related_name="companies")

    email = models.CharField(max_length=256, null=True, blank=True)

    def __str__(self):
        return self.name


class UserType(ScopedPermissionHolder, ScopedModelMixin, models.Model):
    class Meta:
        pass

    name = models.CharField(max_length=128)

    company = models.ForeignKey(
        Company, on_delete=models.CASCADE, related_name="user_types"
    )
    users = models.ManyToManyField(User, blank=True, related_name="user_types")
    is_administrator = models.BooleanField(default=False)

    color = models.CharField(max_length=8, null=True, blank=True)
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    def get_base_scopes(self):
        return [
            create_scope("usertype", self.id),
            create_scope("company", self.company.id, "usertype", self.id),
        ]

    def clean(self):
        if isinstance(self.color, str) and self.color.startswith("#"):
            self.color = self.color[1:]

    def __str__(self):
        return f"{self.name} ({self.company.name})"


class Pet(ScopedModelMixin, models.Model):
    class Meta:
        pass

    user = models.ForeignKey(User, on_delete=models.CASCADE, related_name="pets")

    name = models.CharField(max_length=128)
    age = models.PositiveIntegerField()

    def get_required_scopes(self):
        return [
            create_scope("pet", self.id),
            create_scope("user", self.user.id, "pet", self.id),
        ]

    def __str__(self):
        return self.name



from typing import Tuple, Mapping, Union, Iterable

from graphene import Node
from graphene_django import DjangoObjectType
from graphene_django.types import DjangoObjectTypeOptions
from graphene_django_cud.mutations import DjangoPatchMutation, DjangoDeleteMutation, DjangoBatchCreateMutation, \
    DjangoBatchDeleteMutation, DjangoFilterDeleteMutation
from graphene_django_cud.mutations.create import DjangoCreateMutation
from graphene_django_cud.mutations.delete import DjangoDeleteMutationOptions
from graphene_django_cud.mutations.patch import DjangoPatchMutationOptions
from graphene_django_cud.mutations.update import (
    DjangoUpdateMutationOptions,
    DjangoUpdateMutation,
)
from graphql import GraphQLError

from django_scoped_permissions.util import (
    expand_scopes_from_context,
    create_resolver_from_method,
    create_resolver_from_scopes,
)
from django_scoped_permissions.models import ScopedPermissionHolder, ScopedModelMixin


class ScopedDjangoNodeOptions(DjangoObjectTypeOptions):
    allow_anonymous = False  # type: bool
    node_permissions = None  # type: Tuple[str]
    field_permissions = None  # type: Mapping[str, Union[bool, Tuple[str]]]


class ScopedDjangoNode(DjangoObjectType):
    class Meta:
        abstract = True

    @classmethod
    def __init_subclass_with_meta__(
        cls,
        node_permissions=None,
        field_permissions=None,
        allow_anonymous=False,
        _meta=None,
        **options,
    ):
        if not _meta:
            _meta = ScopedDjangoNodeOptions(cls)

        _meta.allow_anonymous = allow_anonymous
        _meta.node_permissions = node_permissions
        _meta.field_permissions = field_permissions
        interfaces = options.get("interfaces", ()) + (Node,)

        super().__init_subclass_with_meta__(
            _meta=_meta, interfaces=interfaces, **options
        )

        # Great, the class is set up. Now let's add permission guards.
        node_permissions = node_permissions or {}
        field_permissions = field_permissions or {}

        for field, permissions in field_permissions.items():
            if callable(permissions):
                if hasattr(cls, f"resolve_{field}"):
                    continue

                setattr(
                    cls,
                    f"resolve_{field}",
                    create_resolver_from_method(field, permissions),
                )
            elif isinstance(permissions, tuple) or isinstance(permissions, list):
                if hasattr(cls, f"resolve_{field}"):
                    continue

                setattr(
                    cls,
                    f"resolve_{field}",
                    create_resolver_from_scopes(field, permissions),
                )
            elif isinstance(permissions, str):
                if hasattr(cls, f"resolve_{field}"):
                    continue

                setattr(
                    cls,
                    f"resolve_{field}",
                    create_resolver_from_scopes(field, [permissions]),
                )
            else:
                raise ValueError(
                    f"Invalid field type {type(permissions)} given to ScopedDjangoNode for field {field}"
                )

    @classmethod
    def get_node(cls, info, id):
        user = info.context.user
        if not cls._meta.allow_anonymous and user.is_anonymous:
            raise GraphQLError("You are not permitted to view this.")

        if cls._meta.node_permissions:
            if not user.has_scoped_permissions(*cls._meta.node_permissions):
                raise GraphQLError("You are not permitted to view this.")
        else:
            # Try to get object and see if we can get the required scopes from it
            obj = cls._meta.model.objects.get(pk=id)
            if hasattr(obj, "get_base_scopes"):
                if not user.has_scoped_permissions(*obj.get_base_scopes()):
                    raise GraphQLError("You are not permitted to view this.")

        return super().get_node(info, id)


class ScopedDjangoCreateMutation(DjangoCreateMutation):
    class Meta:
        abstract = True

    @classmethod
    def check_permissions(cls, root, info, input) -> None:
        permissions = cls.get_permissions(root, info, input) or []

        if not hasattr(permissions, '__len__') or len(permissions) == 0:
            return

        context = {"context": info.context, "input": input}

        expanded_permissions = expand_scopes_from_context(permissions, context)

        if not isinstance(info.context.user, ScopedPermissionHolder):
            raise GraphQLError("You are not permitted to view this.")

        if not info.context.user.has_scoped_permissions(*expanded_permissions):
            raise GraphQLError("You are not permitted to view this.")

class ScopedDjangoBatchCreateMutation(DjangoBatchCreateMutation):
    class Meta:
        abstract = True

    @classmethod
    def check_permissions(cls, root, info, input) -> None:
        permissions = cls.get_permissions(root, info, input) or []

        if not hasattr(permissions, '__len__') or len(permissions) == 0:
            return

        context = {"context": info.context, "input": input}

        expanded_permissions = expand_scopes_from_context(permissions, context)

        if not isinstance(info.context.user, ScopedPermissionHolder):
            raise GraphQLError("You are not permitted to view this.")

        if not info.context.user.has_scoped_permissions(*expanded_permissions):
            raise GraphQLError("You are not permitted to view this.")


class ScopedDjangoPatchMutationOptions(DjangoPatchMutationOptions):
    permission_action = "update"  # type: str


class ScopedDjangoPatchMutation(DjangoPatchMutation):
    class Meta:
        abstract = True

    @classmethod
    def get_permissions(cls, root, info, input, id, obj) -> Iterable[str]:
        super_permissions = super().get_permissions(root, info, input, id, obj) or []

        if hasattr(super_permissions, '__len__') and len(super_permissions) > 0:
            return super_permissions

        # If we don't have explicit permissions we use some defaults here.
        if cls._meta.permission_action == "":
            return ["{base_scopes}"]
        else:
            return [
                # Double curly brackets escapes them
                f"{{base_scopes}}:{cls._meta.permission_action}"
            ]

    @classmethod
    def check_permissions(cls, root, info, input, id, obj) -> None:
        permissions = cls.get_permissions(root, info, input, id, obj)

        context = {}

        if isinstance(obj, ScopedModelMixin):
            context["base_scopes"] = obj.get_base_scopes()

        context["context"] = info.context
        context["input"] = input
        context["id"] = id
        context["obj"] = obj

        expanded_permissions = expand_scopes_from_context(permissions, context)

        if not isinstance(info.context.user, ScopedPermissionHolder):
            raise GraphQLError("You are not permitted to view this.")

        if not info.context.user.has_scoped_permissions(*expanded_permissions):
            raise GraphQLError("You are not permitted to view this.")

    @classmethod
    def __init_subclass_with_meta__(
        cls, _meta=None, permission_action="update", **options
    ):
        if _meta is None:
            _meta = ScopedDjangoPatchMutationOptions(cls)

        _meta.permission_action = permission_action

        return super().__init_subclass_with_meta__(_meta=_meta, **options)


class ScopedDjangoUpdateMutationOptions(DjangoUpdateMutationOptions):
    permission_action = "update"  # type: str


class ScopedDjangoUpdateMutation(DjangoUpdateMutation):
    class Meta:
        abstract = True

    @classmethod
    def get_permissions(cls, root, info, input, id, obj) -> Iterable[str]:
        super_permissions = super().get_permissions(root, info, input, id, obj) or []

        if hasattr(super_permissions, '__len__') and len(super_permissions) > 0:
            return super_permissions

        # If we don't have explicit permissions we use some defaults here.
        if cls._meta.permission_action == "":
            return ["{base_scopes}"]
        else:
            return [
                # Double curly brackets escapes them
                f"{{base_scopes}}:{cls._meta.permission_action}"
            ]

    @classmethod
    def check_permissions(cls, root, info, input, id, obj) -> None:
        permissions = cls.get_permissions(root, info, input, id, obj) or []

        context = {}

        if isinstance(obj, ScopedModelMixin):
            context["base_scopes"] = obj.get_base_scopes()

        context["context"] = info.context
        context["input"] = input
        context["id"] = id
        context["obj"] = obj

        expanded_permissions = expand_scopes_from_context(permissions, context)

        if not isinstance(info.context.user, ScopedPermissionHolder):
            raise GraphQLError("You are not permitted to view this.")

        if not info.context.user.has_scoped_permissions(*expanded_permissions):
            raise GraphQLError("You are not permitted to view this.")

    @classmethod
    def __init_subclass_with_meta__(
        cls, _meta=None, permission_action="update", **options
    ):
        if _meta is None:
            _meta = ScopedDjangoUpdateMutationOptions(cls)

        _meta.permission_action = permission_action

        return super().__init_subclass_with_meta__(_meta=_meta, **options)


class ScopedDjangoDeleteMutationOptions(DjangoDeleteMutationOptions):
    permission_action = "delete"  # type: str


class ScopedDjangoDeleteMutation(DjangoDeleteMutation):
    class Meta:
        abstract = True

    @classmethod
    def get_permissions(cls, root, info, id, obj) -> Iterable[str]:
        super_permissions = list(super().get_permissions(root, info, id, obj))

        if hasattr(super_permissions, '__len__') and len(super_permissions) > 0:
            return super_permissions

        # If we don't have explicit permissions we use some defaults here.
        if cls._meta.permission_action == "":
            return ["{base_scopes}"]
        else:
            return [
                # Double curly brackets escapes them
                f"{{base_scopes}}:{cls._meta.permission_action}"
            ]

    @classmethod
    def check_permissions(cls, root, info, id, obj) -> None:
        permissions = cls.get_permissions(root, info,  id, obj) or []

        context = {}

        if isinstance(obj, ScopedModelMixin):
            context["base_scopes"] = obj.get_base_scopes()

        context["context"] = info.context
        context["input"] = input
        context["id"] = id
        context["obj"] = obj

        expanded_permissions = expand_scopes_from_context(permissions, context)

        if not isinstance(info.context.user, ScopedPermissionHolder):
            raise GraphQLError("You are not permitted to view this.")

        if not info.context.user.has_scoped_permissions(*expanded_permissions):
            raise GraphQLError("You are not permitted to view this.")

    @classmethod
    def __init_subclass_with_meta__(
        cls, _meta=None, permission_action="delete", **options
    ):
        if _meta is None:
            _meta = ScopedDjangoUpdateMutationOptions(cls)

        _meta.permission_action = permission_action

        return super().__init_subclass_with_meta__(_meta=_meta, **options)


class ScopedDjangoBatchDeleteMutation(DjangoBatchDeleteMutation):
    class Meta:
        abstract = True

    @classmethod
    def check_permissions(cls, root, info, input) -> None:
        permissions = list(cls.get_permissions(root, info, input)) or []

        if not hasattr(permissions, '__len__') or len(permissions) == 0:
            return

        context = {"context": info.context, "input": input}

        expanded_permissions = expand_scopes_from_context(permissions, context)

        if not isinstance(info.context.user, ScopedPermissionHolder):
            raise GraphQLError("You are not permitted to view this.")

        if not info.context.user.has_scoped_permissions(*expanded_permissions):
            raise GraphQLError("You are not permitted to view this.")


class ScopedDjangoFilterDeleteMutation(DjangoFilterDeleteMutation):
    class Meta:
        abstract = True

    @classmethod
    def check_permissions(cls, root, info, input) -> None:
        permissions = list(cls.get_permissions(root, info, input)) or []

        if not hasattr(permissions, '__len__') or len(permissions) == 0:
            return

        context = {"context": info.context, "input": input}

        expanded_permissions = expand_scopes_from_context(permissions, context)

        if not isinstance(info.context.user, ScopedPermissionHolder):
            raise GraphQLError("You are not permitted to view this.")

        if not info.context.user.has_scoped_permissions(*expanded_permissions):
            raise GraphQLError("You are not permitted to view this.")



import itertools
import re
from typing import Mapping, Iterable
from django_scoped_permissions.models import ScopedModel

from pydash import get


def create_resolver_from_method(field_name, method):
    from graphql import GraphQLError
    def resolver(object, info, **args):
        if not method(object, info, **args):
            raise GraphQLError("You are not permitted to view this.")

        return getattr(object, field_name)

    return resolver


def create_resolver_from_scopes(field_name: str, scopes: [str]):
    from graphql import GraphQLError
    def resolver(object, info, **args):

        user = info.context.user

        final_scopes = []
        format_variables = []

        field_value = getattr(object, field_name, None)

        field_value_is_scoped_model = isinstance(field_value, ScopedModel)
        object_is_scoped_model = isinstance(object, ScopedModel)

        field_value_base_scopes = [""]
        object_base_scopes = [""]

        expansion_map = {}

        if field_value_is_scoped_model:
            field_value_base_scopes = field_value.get_base_scopes()

        if object_is_scoped_model:
            object_base_scopes = object.get_base_scopes()

        expansion_map["base_scopes"] = field_value_base_scopes
        expansion_map["field_scopes"] = object_base_scopes

        final_scopes = []
        for scope in scopes:
            if "{base_scopes}" and "{field_scopes}" in scope:
                product = itertools.product(object_base_scopes, field_value_base_scopes)

                for combination in product:
                    final_scopes.append(
                        scope.format(
                            base_scopes=combination[0], field_scopes=combination[1]
                        )
                    )
            elif "{base_scopes}" in scope:
                for base_scope in object_base_scopes:
                    final_scopes.append(scope.format(base_scopes=base_scope))
            elif "{field_scopes}" in scope:
                for field_scope in field_value_base_scopes:
                    final_scopes.append(scope.format(field_scopes=base_scope))
            else:
                final_scopes.append(scope)

        if not user.has_any_scoped_permissions(*final_scopes):
            raise GraphQLError("You are not permitted to view this.")

        return field_value

    return resolver


def expand_scopes(
    scopes: Iterable[str], expansion_map: Mapping[str, Iterable[str]] = None
) -> Iterable[str]:
    if expansion_map is None or len(list(expansion_map.values())) == 0:
        return scopes

    # First get all keys of the expansion map.
    keys = list(expansion_map.keys())

    values = expansion_map.values()

    # Now create the cartesian product of all expansion values
    product = itertools.product(*values)

    final_scopes = set()

    # Brilliant, so now we have all the keys, and all the possible permutations of their
    # values. We can now create all the final interpolated scopes
    for permutation in product:
        # Construct the expansion map for this permutation
        permutation_expansion_map = {keys[i]: permutation[i] for i in range(len(keys))}

        for scope in scopes:
            final_scopes.add(scope.format(**permutation_expansion_map))

    return list(final_scopes)


scope_variable_regex = re.compile("{[^{}]+}")


def expand_scopes_from_context(scopes: Iterable[str], context):
    """
    Takes a context object, and expands all scopes. The context object may contain nested dictionaries,
    or other object types.

    To extract the relevant fields from the context, we use pydash.get.

    This method is used to inject appropriate variables into the scope strings. Suppose for instance
    that we have the scope "facilities:{input.facility}:create". The idea here is that we should take
    the input as a context-parameter, and inject the "facility" field into the scope.

    Before sending the variables and scopes to the "expand_scopes" helper method, we translate all
    dots to double underscores. This is to please the str.format method, which attempts to do
    magic itself when getting variables with dots. Using variables with double underscores
    directly in the scopes sent to this method should be unproblematic, but may lead to undefined behaviour.

    Also note that some variables may have multiple target values. This is the case for the magic variable
    {base_scopes}, which can be used for mutations and queries where a singular object of type ScopedModel
    is supplied. This is the case for DjangoPatchMutation, DjangoUpdateMutation and DjangoScopedNode.

    :param scopes:
    :param context:
    :return:
    """

    # First we extract all the variable strings we need to attend to. We prefetch them here, so we
    # can extract them from the context
    variable_strings = []
    new_scopes = []

    for scope in scopes:
        matches = scope_variable_regex.findall(scope)
        for match in matches:
            variable_strings.append(match[1:-1])

        # We translate dots to double underscores here. This is to please
        # the str.format function, as it itself tries to do some magic
        # when receiving dotted variables.
        new_scopes.append(scope.replace(".", "__"))

    extracted_context_values = {
        # We replace the double underscores here as well, so the lookups match
        variable.replace(".", "__"): _overload_context_variable(get(context, variable))
        for variable in variable_strings
    }

    return expand_scopes(new_scopes, extracted_context_values)


def _overload_context_variable(variable):
    """
    This helper method basically ensures a context variable is wrapped in a list
    if it is not a list itself.

    :param variable:
    :return:
    """
    if isinstance(variable, list):
        return variable
    return [variable]



from addict import Dict
from django.core.exceptions import PermissionDenied
from django.test import TestCase

from django_scoped_permissions.decorators import gql_has_scoped_permissions
from django_scoped_permissions.models import ScopedPermission
from django_scoped_permissions.tests.factories import UserFactory, CompanyFactory


class TestSqlHasScopedPermissions(TestCase):
    def test__simple_scope_check_with_permission__permits(self):

        @gql_has_scoped_permissions("scope1:scope2")
        def wrapper_method(self, info):
            pass

        user = UserFactory.create()
        perm = ScopedPermission.objects.create(scope="scope1:scope2")
        user.scoped_permissions.add(perm)
        info = Dict()
        info.context.user = user
        wrapper_method(None, info)

    def test__simple_scope_check_without_permission__fails(self):

        @gql_has_scoped_permissions("scope1:scope2")
        def wrapper_method(self, info):
            pass

        user = UserFactory.create()
        perm = ScopedPermission.objects.create(scope="scope3")
        user.scoped_permissions.add(perm)
        info = Dict()
        info.context.user = user
        with self.assertRaises(PermissionDenied):
            wrapper_method(None, info)


    def test_expand_from_context__check_with_organization__succeeds(self):

        @gql_has_scoped_permissions("company:{context.company.id}:read")
        def wrapper_method(self, info):
            pass

        user = UserFactory.create()
        perm = ScopedPermission.objects.create(scope="company:1")
        user.scoped_permissions.add(perm)
        info = Dict()
        info.context.user = user
        info.context.company = CompanyFactory.create()
        wrapper_method(None, info)
